from maaml.Datasets.UAH_dataset import time_series
from maaml import preprocessing
from maaml import machine_learning
from maaml.functions.Datasets import time_series
from maaml.functions import preprocessing
from maaml.functions import machine_learning
from maaml.functions import deep_learning
